/* 
 * XmlTclXml.c -- (c) 2010 Cyril BARABTO 
 *
 *	This file contains code to implement the "packer"
 *	XML file manager for Tcl/Tk.
 *
 * libxml (c) 1999-2010 Daniel Veillard
 * tcl/tk (c) 1990-2010 The Regents of the University of California.
 * 	  (c) 1994-2010 Sun Microsystems, Inc.
 *
 * DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include <libxml/HTMLparser.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "XmlTclInit.h"

/*
 *--------------------------------------------------------------
 * 
 * declare type def xmltcldoc and xmltclnode.
 *
 *--------------------------------------------------------------
 */

/* xmltcldoc */
typedef struct {

	/* tcl part */
	Tcl_Interp 	*interp;		/* interpreter associated with widget */
 	Tcl_Command 	widgetCmd;		/* token for xml doc widget command */
	char		*filename;		/* file name (root only) */
	char 		*name;			/* widget name 	*/
	char		*args;			/* args string for another using */ 
	char		load;			/* 0 if not load file else 1 */
	char		html;			/* 0 if not html doc  else 1 */

	/* lib xml part */
	xmlDocPtr	doc;			/* xml doc */
	
} XmlTclDoc;

/* xmltclnode */
typedef struct {
	
	/* tcl part */
	Tcl_Interp 	*interp;		/* interpreter associated with widget */
 	Tcl_Command 	widgetCmd;		/* token for xml node widget command 	*/
	char 		*name;			/* widget name 	*/
	char 		*parent;		/* widget parent name 	*/
	char		*args;			/* args string for another using */ 

	/* lib xml part */
	xmlNodePtr	node;			/* current node */

}  XmlTclNode;

/*
 *--------------------------------------------------------------
 * 
 * define commands and options index.
 *
 *--------------------------------------------------------------
 */

	
#define	XMLTCL_FILE		20	/* options or args */
#define XMLTCL_CONTENT		30
#define XMLTCL_CONFIGURE	100	/* commands */
#define XMLTCL_CGET		110
#define XMLTCL_ADD		120
#define XMLTCL_SET		130
#define XMLTCL_GET		140
#define XMLTCL_SAVE		150
#define XMLTCL_DELETE		160
#define XMLTCL_INFO		170
#define XMLTCL_ROOT		1000	/* sub commands */
#define XMLTCL_NODE		1100
#define XMLTCL_NODESLIST	1110
#define XMLTCL_THISNODE		1120	/* + october 2008 */
#define XMLTCL_PROPERTY		1200	
#define XMLTCL_PROPERTIES	1210
#define XMLTCL_PROPERTIESLIST	1220
#define XMLTCL_TAGNAME		1300
#define XMLTCL_NAME		1400
#define XMLTCL_PARENT		1500
#define XMLTCL_HTML		2000	/* switch */

/*
 *--------------------------------------------------------------
 * 
 * define equ options/commands index with options/commands reels names
 *
 *--------------------------------------------------------------
 */

XmlTcl_Equ xmltclEqu[] = {
	{"-file", 	XMLTCL_FILE		},
	{"-content",	XMLTCL_CONTENT		},
	{"configure",	XMLTCL_CONFIGURE	},
	{"cget",	XMLTCL_CGET		},
	{"root",	XMLTCL_ROOT		},
	{"add",		XMLTCL_ADD		},
	{"set",		XMLTCL_SET		},
	{"get",		XMLTCL_GET		},
	{"save",	XMLTCL_SAVE		},
	{"delete",	XMLTCL_DELETE		},
	{"info",	XMLTCL_INFO		},
	{"node",	XMLTCL_NODE		},
	{"nodeslist",	XMLTCL_NODESLIST	},
	{"thisnode",	XMLTCL_THISNODE		},	//+ october 2008
	{"content",	XMLTCL_CONTENT		},
	{"property",	XMLTCL_PROPERTY		},
	{"properties",	XMLTCL_PROPERTIES	},
	{"propertieslist", XMLTCL_PROPERTIESLIST },
	{"tagname",	XMLTCL_TAGNAME		},
	{"name",	XMLTCL_NAME		},
	{"parent",	XMLTCL_PARENT		},
	{"-html",	XMLTCL_HTML		},
	{0, 0}
};	

/*
 *--------------------------------------------------------------
 * 
 * declare prototypes widgets about xml commands and procs
 *
 *--------------------------------------------------------------
 */


static int  	XmlDocWidgetCmd(ClientData, Tcl_Interp *, int , char **); 	/* widget cmd xml */
static void 	XmlDocWidgetDelete(ClientData); 				/* proc kill xml */
static int  	XmlDocWidgetConfigure(Tcl_Interp *, XmlTclDoc *, int , char **);/* save option and config xml widget */
static XmlTclNode * XmlNodeWidgetCreate(Tcl_Interp *, char *, xmlNodePtr);    	/* widget cmd add node */
static int  	XmlNodeWidgetCmd(ClientData, Tcl_Interp *, int , char **); 	/* widget cmd xml node*/
static void 	XmlNodeWidgetDelete(ClientData); 				/* proc kill xml node*/
static int  	XmlNodeWidgetSet(Tcl_Interp *, XmlTclNode *, int , char **);	/* save option and config xml node widget */
static int	XmlTreeParseWidget(Tcl_Interp *, char *, xmlNodePtr);  		/* add all nodes wigets for a xml tree in memory */
static int	XmlGetNodeN(xmlNodePtr, xmlNodePtr);

/*
 *--------------------------------------------------------------
 *
 * XmlTcl_XmlCmd --
 *
 *	This procedure is invoked to process the "xml" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */
int 
XmlTcl_XmlCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int   	argc;		/* Number of arguments. */
    char 	**argv;		/* Argument strings. */
    {
	XmlTclDoc	*xmltcldocPtr;	/* current widget xml pointer*/
	char	*tamponStr	= NULL;	/* tampon command string */

	if ((argc <2 ) || (argv[1][0]!='.')) {
		 sprintf(interp->result, "%s ?pathName? ?arg? ?value? ... ?args? ?value?!", argv[0]);
		 return TCL_ERROR ;
	};

	/* create and init new xml option register  */  	
	xmltcldocPtr 		= (XmlTclDoc *) ckalloc(sizeof(XmlTclDoc));
 	xmltcldocPtr->interp 	= interp;
  	xmltcldocPtr->name 	= (char *) malloc (strlen(argv[1])+1);
	strncpy(xmltcldocPtr->name, argv[1], strlen(argv[1])+1);
	/* NEW CMD */
	xmltcldocPtr->widgetCmd 	= Tcl_CreateCommand(interp, xmltcldocPtr->name,
						XmlDocWidgetCmd, (ClientData) xmltcldocPtr, 
						XmlDocWidgetDelete);
	
	/* flags */
	xmltcldocPtr->load	= 0;
	xmltcldocPtr->html	= 0;

	/* file name */
	xmltcldocPtr->filename = (char *) malloc (512);
	xmltcldocPtr->filename[0]=0;

	/* another args */
	xmltcldocPtr->args	= (char *) NULL;
	
	/* CONFIGURE XmlDoc WIDGET */
	if (XmlDocWidgetConfigure(interp, xmltcldocPtr, argc-2, argv+2)) return TCL_ERROR;

	/* default file name */
	if (xmltcldocPtr->filename[0]==0) {
		if (!xmltcldocPtr->html) {
				xmltcldocPtr->filename  = (char *) malloc (strlen(argv[1])+1);
				sprintf(xmltcldocPtr->filename, "%s.xml", argv[1]+1);
			} else {
				xmltcldocPtr->filename  = (char *) malloc (strlen(argv[1])+1);
				sprintf(xmltcldocPtr->filename, "%s.html", argv[1]+1);
		};
	};

	/* create a new doc if file not open */ 
	if (!xmltcldocPtr->load) {
		if (!xmltcldocPtr->html) 
				xmltcldocPtr->doc 	= xmlNewDoc("0.1");
			else {
				xmltcldocPtr->doc 	= (xmlDocPtr) htmlNewDoc(NULL, NULL);
				XmlNodeWidgetCreate(interp, argv[1],
					xmltcldocPtr->doc->children=xmlNewNode(NULL, "html"));
				XmlNodeWidgetCreate(interp, 
					(char *) strcat(argv[1],".html"),
					xmlNewChild(xmltcldocPtr->doc->children,
					NULL, "head", NULL));
				XmlNodeWidgetCreate(interp, 
					argv[1],
					xmlNewChild(xmltcldocPtr->doc->children,
					NULL, "body", NULL));
			};
	};

	/* return widget name for tcl result */
	interp->result 	= xmltcldocPtr->name;
	
	/* OK */
	return TCL_OK;
};

/*
 *--------------------------------------------------------------
 *
 * XmlDocWidgetDelete --
 *
 *	called by widget command when destroy
 *
 *--------------------------------------------------------------
 */
static void
XmlDocWidgetDelete(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
	XmlTclDoc	*xmltcldocPtr = (XmlTclDoc *) clientData;	/* find xmltcldocPtr */
	free(xmltcldocPtr);
};

/*
 *--------------------------------------------------------------
 *
 * XmlDocWidgetCmd --
 * 
 * 	called by "?xmlPathName?" widget
 *
 * Results:
 *	A standard Tcl result.
 *
 *--------------------------------------------------------------
 */
static int
XmlDocWidgetCmd(clientData, interp, argc, argv)
    ClientData  clientData;	/* Main window. */
    Tcl_Interp 	*interp;	/* Current interpreter. */
    int 	argc;		/* Number of arguments. */
    char 	**argv;		/* Argument strings. */
{
   	XmlTclDoc *xmltcldocPtr = (XmlTclDoc *) clientData;	/* find XmlTclDoc */
	int index 		= 0;				/* index of command */
	int	code_erreur	= TCL_ERROR;			/* error */	
	
	/* test args number */
	if (argc<2) { 
		sprintf(interp->result, "%s ?command? ?arg?.....?arg?!", argv[0]);
		return TCL_ERROR;
	};
 
	/* indentify command */
	index = XmlTcl_GetEqu(argv[1], xmltclEqu);
	sprintf(interp->result, "\"%s\" command \"%s\" syntax error or not exist!", argv[0], argv[1]);
	switch (index) {
		case XMLTCL_ROOT :
				if (argv[2]==NULL) break;
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_NODE :
						if (argv[3]==NULL) break;
						XmlNodeWidgetSet(interp,
							XmlNodeWidgetCreate(interp,
								argv[0],
								xmltcldocPtr->doc->children=xmlNewNode(NULL,
								argv[3]+1)),
							argc-4, argv+4);
						code_erreur = TCL_OK;
						break;
				};
				break ;
		case XMLTCL_CGET :
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_FILE :
						interp->result=xmltcldocPtr->filename;
						code_erreur = TCL_OK; 
						break;
				};
				break;
		case XMLTCL_INFO :
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_NAME :
						interp->result=xmltcldocPtr->name;	
						code_erreur = TCL_OK; 
						break;
				};
				break;
		case XMLTCL_CONFIGURE : 
				/* CONFIGURE XmlDoc WIDGET */
				code_erreur = 	XmlDocWidgetConfigure(interp, xmltcldocPtr, argc-2, argv+2);
				break ;
		case XMLTCL_GET	:
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_ROOT :
						sprintf(interp->result, "%s.%s",  xmltcldocPtr->name, xmltcldocPtr->doc->children->name);
						code_erreur = TCL_OK;
						break;
				};
				break;
		case XMLTCL_SAVE : 
				/* ReCONFIGURE XmlDoc WIDGET */
				xmltcldocPtr->load = 1;
				code_erreur = 	XmlDocWidgetConfigure(interp, xmltcldocPtr, argc-2, argv+2);
				/* SAVE */
				if (!xmltcldocPtr->html) 
						xmlSaveFile(xmltcldocPtr->filename, xmltcldocPtr->doc);
					else
						htmlSaveFile(xmltcldocPtr->filename, xmltcldocPtr->doc);
				xmltcldocPtr->load = 0;	
				break ;
	};

	/* return ERROR CODE */
	return code_erreur;
};

/*
 *--------------------------------------------------------------
 *
 * XmlDocWidgetConfigure --
 *
 * 	store configuration and configure xml widget
 *
 * Results:
 *	A standard Tcl result.
 * 
 *--------------------------------------------------------------
 */
static int
XmlDocWidgetConfigure(interp,  xmltcldocPtr, argc, argv)
    Tcl_Interp  *interp;         	/* Used for error reporting. */
    XmlTclDoc 	*xmltcldocPtr;   	/* Information about widget. */
    int 	argc;                   /* Number of valid entries in argv. */
    char 	**argv;                	/* Arguments. */

{    
	int i			= 0;			/* index of args */
	int index 		= 0;			/* index of option */
	char *args;					/* string of not used argv */	

	/* string of args */
	args 	= (char *) malloc (1024);
	args[0] = 0x00;
	
	/* error message */
	sprintf(interp->result, "options config. syntax error!", argv[0]);

	/* stock values in register */
	for  ( i = 0 ; i<argc ; i++ ) {
		index = XmlTcl_GetEqu(argv[i], xmltclEqu);
		switch (index) {
			case XMLTCL_HTML	: 
				xmltcldocPtr->html 	= 1;
				break;
 			case XMLTCL_FILE	:
				if (argv[i+1]==NULL) return TCL_ERROR; 
				strcpy(xmltcldocPtr->filename, argv[i+1]);
				/* LOAD FILE */
				if (!xmltcldocPtr->load) {
					if (!xmltcldocPtr->html) {
							/* PARSE XML FILE */
							xmltcldocPtr->doc = xmlParseFile(xmltcldocPtr->filename);
							if (xmltcldocPtr->doc==0)  { 
								sprintf(interp->result, "\"%s\" is not xml file or not exist!", xmltcldocPtr->filename); 
								return TCL_ERROR;
							}
						} else {
							/* PARSE HTML FILE */
							xmltcldocPtr->doc = (xmlDocPtr) htmlParseFile(xmltcldocPtr->filename, NULL);
							if (xmltcldocPtr->doc==0)  { 
								sprintf(interp->result, "\"%s\" is not html file or not exist!", xmltcldocPtr->filename); 
								return TCL_ERROR;
							}
					};
					if (XmlTreeParseWidget(interp, xmltcldocPtr->name, xmlDocGetRootElement(xmltcldocPtr->doc))) return TCL_ERROR;
					xmltcldocPtr->load = 1;
					
				};
				i++;
				break ;
			default 		: 
				sprintf(args, "%s %s %s", args, argv[i-1], argv[i]); 
				break;
		};
	
	};

	
	/* another args not used */ 
	free(xmltcldocPtr->args);
	xmltcldocPtr->args 	=	(char *) malloc (strlen(args)+1);
	strncpy(xmltcldocPtr->args,args, strlen(args)+1);	
	free (args);

	/* return widget name for tcl result */
	interp->result = xmltcldocPtr->name;	

	/* ok */
	return TCL_OK ;
};

/*
 *--------------------------------------------------------------
 *
 * XmlNodeWidgetCreate --
 *
 *	called by : ?xmlpathName? add node
 *
 * Results:
 *	XmlTclNode result.
 * 
 *--------------------------------------------------------------
 */
static XmlTclNode *
XmlNodeWidgetCreate(interp, parentwidgetname, nodePtr)
    Tcl_Interp  *interp;         	/* Used for error reporting. */
    char 	*parentwidgetname;
    xmlNodePtr	nodePtr;
{    
	XmlTclNode	*xmltclnodePtr;
	Tcl_CmdInfo	*cmdinfoPtr;
	char		tclscriptStr[512];
	int		index = 0;
	
	/* create and init new xml option register  */  	
	xmltclnodePtr 		= (XmlTclNode *) ckalloc (sizeof(XmlTclNode));
	xmltclnodePtr->interp 	= interp;
  	xmltclnodePtr->name 	= (char *) malloc (strlen(parentwidgetname)+strlen(nodePtr->name)+2);
	sprintf(xmltclnodePtr->name, "%s.%s", parentwidgetname, nodePtr->name); 
	/* NEW CMD */
	sprintf(tclscriptStr,"catch { %s info tagname }", xmltclnodePtr->name); 
	Tcl_Eval(interp, tclscriptStr);
	while (Tcl_GetStringResult(interp)[0]=='0') {
		index++;
		free(xmltclnodePtr->name);
		xmltclnodePtr->name = (char *) malloc (strlen(parentwidgetname)+strlen(nodePtr->name)+4+index);
		sprintf(xmltclnodePtr->name, "%s.%s(%d)", parentwidgetname, nodePtr->name, index); 
		sprintf(tclscriptStr,"catch { %s info tagname }", xmltclnodePtr->name); 
		Tcl_Eval(interp, tclscriptStr);
	} 


	xmltclnodePtr->widgetCmd 	= Tcl_CreateCommand(interp, xmltclnodePtr->name,
						XmlNodeWidgetCmd, (ClientData) xmltclnodePtr, 
						XmlNodeWidgetDelete);
	/* client node */
	xmltclnodePtr->node	= nodePtr;

	/* parent */
	xmltclnodePtr->parent 		= (char *) malloc (strlen(parentwidgetname)+1);
	strcpy(xmltclnodePtr->parent, parentwidgetname);

	/* another args */
	xmltclnodePtr->args	= (char *) NULL;

	/* return widget name for tcl result */
	interp->result=xmltclnodePtr->name;

	/* return pointer */
	return xmltclnodePtr;
};

/*
 *--------------------------------------------------------------
 *
 * XmlNodeWidgetDelete --
 *
 *	called by widget command when destroy
 *
 *--------------------------------------------------------------
 */
static void
XmlNodeWidgetDelete(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
	XmlTclNode	*xmltclnodePtr = (XmlTclNode *) clientData;	/* find xmltclnodePtr */
	free(xmltclnodePtr);
};

/*
 *--------------------------------------------------------------
 *
 * XmlNodeWidgetCmd --
 * 
 * 	called by widget ?xmlNodepathName?
 *
 * Results:
 *	A standard Tcl result.
 *
 *--------------------------------------------------------------
 */
static int
XmlNodeWidgetCmd(clientData, interp, argc, argv)
    ClientData  clientData;	/* Main window. */
    Tcl_Interp 	*interp;	/* Current interpreter. */
    int 	argc;		/* Number of arguments. */
    char 	**argv;		/* Argument strings. */
{
   	XmlTclNode 	*xmltclnodePtr = (XmlTclNode *) clientData;	/* find XmlTclNode */
	xmlNodePtr 	xmlnodePtr;
	xmlAttrPtr  	xmlattrPtr;
	Tcl_Obj		*listPtr;
	char		tclscriptStr[512];

	int index 		= 0;				/* index of command */
	int	code_erreur	= TCL_ERROR;			/* error */	

	/* test args number */
	if (argc<2) { 
		sprintf(interp->result, "%s ?command? ?arg?.....?arg?!", argv[0]);
		return TCL_ERROR;
	};

	/* init a new list for some cmd */
	listPtr = Tcl_NewListObj(0, (Tcl_Obj **) NULL);
 	
	/* indentify command */
	index = XmlTcl_GetEqu(argv[1], xmltclEqu);
	sprintf(interp->result, "\"%s\" command \"%s\" syntax error or options not exist!", argv[0], argv[1]);
	switch (index) {
		case XMLTCL_ADD	:
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_NODE :
						if (argv[3]==NULL) break;
						XmlNodeWidgetSet(interp,
							XmlNodeWidgetCreate(interp,
								argv[0],
								xmlNewChild(xmltclnodePtr->node,
									NULL, argv[3]+1, NULL)),
							argc-4, argv+4);
						code_erreur = TCL_OK;
						break;
					case XMLTCL_PROPERTY :
					case XMLTCL_PROPERTIES :
						if (argv[3]==NULL) break; 
						index = 0;
						while (argv[3+index]!=NULL) {
							if (xmlNewProp(xmltclnodePtr->node, argv[3+index]+1, argv[4+index])==NULL)
								return TCL_ERROR;
							index+=2;
						};
						code_erreur = TCL_OK;
						break;
				};
				break ;
		case XMLTCL_DELETE	:
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					// + thisnode october 2008
					case XMLTCL_THISNODE :
						xmlnodePtr = xmltclnodePtr->node;
						xmlnodePtr->prev->next = xmlnodePtr->next;
						xmltclnodePtr->node = xmlnodePtr->next;
						free(xmlnodePtr);
						//sprintf(tclscriptStr, "%s%s", argv[0], argv[3]);
						//code_erreur = Tcl_DeleteCommand(interp, tclscriptStr);
						code_erreur = TCL_OK;
						break;
					case XMLTCL_NODE :
						if (argv[3]==NULL) break; 
						xmlnodePtr = xmltclnodePtr->node->children;
						while (strcmp(xmlnodePtr->name, argv[3]+1)!=0) 
							if ((xmlnodePtr = xmlnodePtr->next)==NULL) return TCL_ERROR;
						if (xmlnodePtr!=xmltclnodePtr->node->children) 
									xmlnodePtr->prev->next = xmlnodePtr->next;
								else
									xmltclnodePtr->node->children = xmlnodePtr->next;
						xmlnodePtr->prev = NULL;
						xmlnodePtr->next = NULL;
						free(xmlnodePtr);
						sprintf(tclscriptStr, "%s%s", argv[0], argv[3]);
						code_erreur = Tcl_DeleteCommand(interp, tclscriptStr);
						break;
					case XMLTCL_PROPERTY :
					case XMLTCL_PROPERTIES :
						if (argv[3]==NULL) break; 
						index = 0;
						while (argv[3+index]!=NULL) {
							if ((xmlattrPtr = xmlHasProp(xmltclnodePtr->node, argv[3+index]+1))==NULL)
								return TCL_ERROR;
							if (xmlattrPtr!=xmltclnodePtr->node->properties) 
									xmlattrPtr->prev->next = xmlattrPtr->next;
								else
									xmltclnodePtr->node->properties = xmlattrPtr->next;
							xmlattrPtr->prev = NULL;
							xmlattrPtr->next = NULL;
							free(xmlattrPtr);
							index++;
						};
						code_erreur = TCL_OK;
						break;
				};
				break ;
		case XMLTCL_GET	:
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_CONTENT :
						interp->result = xmlNodeGetContent(xmltclnodePtr->node);
						code_erreur = TCL_OK;
						break;
					case XMLTCL_PROPERTY :
						if (argv[3]==NULL) break; 
						if (xmlGetProp(xmltclnodePtr->node, argv[3]+1)==NULL) return TCL_ERROR;
						interp->result=xmlGetProp(xmltclnodePtr->node, argv[3]+1);
						code_erreur = TCL_OK;
						break;
					case XMLTCL_NODESLIST :
						xmlnodePtr = xmltclnodePtr->node->children;
						while (xmlnodePtr!=NULL) {
							if (index = XmlGetNodeN(xmltclnodePtr->node->children, xmlnodePtr)) {
 									sprintf(interp->result, "%s.%s(%d)", 
										xmltclnodePtr->name,
										xmlnodePtr->name, index);
								} else {
									sprintf(interp->result, "%s.%s",
										xmltclnodePtr->name,
										xmlnodePtr->name);
							};
							Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(interp->result, -1));
							xmlnodePtr = xmlnodePtr->next;
						};
						Tcl_SetObjResult(interp, listPtr);
						code_erreur = TCL_OK;
						break;
					case XMLTCL_PROPERTIESLIST :
						strcpy(interp->result, "{");
						xmlattrPtr = xmltclnodePtr->node->properties;
						while (xmlattrPtr!=NULL) {
							sprintf(interp->result, "-%s", xmlattrPtr->name);
							Tcl_ListObjAppendElement(interp, listPtr, Tcl_NewStringObj(interp->result, -1));
							xmlattrPtr = xmlattrPtr->next;
						};
						Tcl_SetObjResult(interp, listPtr);
						code_erreur = TCL_OK;
						break;
				};
				break ;
		case XMLTCL_INFO :
				if (argv[2]==NULL) break; 
				index = XmlTcl_GetEqu(argv[2], xmltclEqu);
				switch (index) {
					case XMLTCL_PARENT :
						interp->result = xmltclnodePtr->parent;	
						code_erreur = TCL_OK; 
						break;
					case XMLTCL_NAME :
						interp->result = xmltclnodePtr->name;	
						code_erreur = TCL_OK; 
						break;
					case XMLTCL_TAGNAME :
						interp->result = (char *) xmltclnodePtr->node->name;	
						code_erreur = TCL_OK; 
						break;
				};
				break;
		case XMLTCL_SET	:
				if (argv[2]==NULL) break; 
				XmlNodeWidgetSet(interp, xmltclnodePtr, argc-2, argv+2);
				code_erreur = TCL_OK;
				break ;
	};

	/* return error */				
	return code_erreur;
};
/*
 *--------------------------------------------------------------
 *
 * XmlNodeWidgetSet --
 *
 * 	store configuration and configure xml widget
 *
 * Results:
 *	A standard Tcl result.
 * 
 *--------------------------------------------------------------
 */
static int
XmlNodeWidgetSet(interp,  xmltclnodePtr, argc, argv)
    Tcl_Interp  *interp;         	/* Used for error reporting. */
    XmlTclNode 	*xmltclnodePtr;   	/* Information about widget. */
    int 	argc;                   /* Number of valid entries in argv. */
    char 	**argv;                	/* Arguments. */

{    
	int i			= 0;			/* index of args */
	int index 		= 0;			/* index of option */
	char *args;					/* string of not used argv */	

	/* string of args */
	args 	= (char *) malloc (1024);
	args[0] = 0x00;

	/* stock values in register */
	for  ( i = 0 ; i<argc ; i++ ) {
		index = XmlTcl_GetEqu(argv[i], xmltclEqu);
		i++;
		if (argv[i]==NULL) {	 
				sprintf(interp->result, "%s configure ?arg? ?value? ... ?args? ?value?!", argv[0]);
	 		return TCL_ERROR ;
		};
		switch (index) {
			case XMLTCL_CONTENT :
				xmlNodeSetContent(xmltclnodePtr->node, argv[i]);
				break;
			case XMLTCL_PROPERTY :
				xmlSetProp(xmltclnodePtr->node, argv[i]+1, argv[i+1]);
				i++;
				break;
			default 		: 
				sprintf(args, "%s %s %s", args, argv[i-1], argv[i]); 
				break;
		};
	
	};

	
	/* another args not used */ 
	free(xmltclnodePtr->args);
	xmltclnodePtr->args 	=	(char *) malloc (strlen(args)+1);
	strncpy(xmltclnodePtr->args,args, strlen(args)+1);	
	free (args);

	/* return widget name for tcl result */
	interp->result = xmltclnodePtr->name;	

	/* ok */
	return TCL_OK ;
};

/*
 *--------------------------------------------------------------
 *
 * XmlTreeParseWidget --
 *
 *	for a xml tree in memory associate all widgets nodes
 *
 * Results:
 *	tcl result.
 * 
 *--------------------------------------------------------------
 */
static int
XmlTreeParseWidget(interp, parentwidgetpath, nodePtr)
    Tcl_Interp  *interp;         	/* Used for error reporting. */
    char *parentwidgetpath;	
    xmlNodePtr	nodePtr;
{    
	XmlTclNode	*xmltclnodePtr;
	if (nodePtr==NULL) return TCL_OK;
	xmltclnodePtr = XmlNodeWidgetCreate(interp, parentwidgetpath, nodePtr);
	if (xmltclnodePtr==NULL) return TCL_ERROR;
	XmlTreeParseWidget(interp, xmltclnodePtr->name, nodePtr->children);
	return XmlTreeParseWidget(interp, parentwidgetpath, nodePtr->next);
};

/*
 *--------------------------------------------------------------
 *
 * XmlGetNodeN --
 *
 *	count same name nodes number
 *
 * Results:
 *	0 if node not exits or if node is first.
 *	else return n (number of node which have same name) 
 * 
 *--------------------------------------------------------------
 */
static int
XmlGetNodeN(nodebasePtr, nodesearchedPtr)
    xmlNodePtr nodebasePtr;	
    xmlNodePtr nodesearchedPtr;
{    
	static index = 0;	/* index of tag which have same mane */
	int result;		/* for return */
	/* test node */
	if ((nodebasePtr==nodesearchedPtr) || (nodebasePtr==NULL)) {
		result = index; index = 0;
		return result;
	}
	if (strcmp(nodebasePtr->name, nodesearchedPtr->name)==0) index++;
	return XmlGetNodeN(nodebasePtr->next, nodesearchedPtr);
};




